#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

from collections import OrderedDict
from .snp_count import (
    get_supporting_reads,
    get_fraction,
    get_supporting_reads_single_region,
)
from .gmm import Gmm
from .bin_count import get_normed_depth
from .copy_number_call import (
    call_cn_snp,
    call_cn_var,
    call_cn_var_homo,
    get_called_variants,
)
from .utilities import open_alignment_file


class GeneCaller:
    """
    A class for calling variant in paralogs
    """
    def __init__(self):
        self.bamfile = None
        self.call_parameters = None
        self.reference = None
        self.threads = None
        self.normalized_depth = None
        self.gmm_call = None
        self.total_cn = None
        self.raw_count = OrderedDict()
        self.gene1_read_count = []
        self.gene2_read_count = []
        self.gene1_fraction = []
        self.raw_gene1_cn = []
        self.var_homo_alt = []
        self.var_homo_ref = []
        self.var_alt = []
        self.var_ref = []
        self.var_list = []
        self.cn_call_snp = []
        self.total_callset = []

    def set_par(self, bamfile, call_parameters, threads=None, reference=None):
        """Establish parameter values"""
        self.bamfile = bamfile
        self.call_parameters = call_parameters
        self.threads = threads
        self.reference = reference
        self.var_list = self.call_parameters.var_list

    def normalize_depth(self, gc_correct=True):
        """Count reads and perform depth normalization"""
        self.normalized_depth = get_normed_depth(
            self.bamfile,
            self.call_parameters.region_dic,
            self.threads,
            self.reference,
            gc_correct=gc_correct
        )

    def call_cn(self, region_name):
        """Call CN for regions of interest"""
        gmm = Gmm()
        gmm_parameter = self.call_parameters.gmm_parameter
        gmm.set_gmm_par(gmm_parameter, region_name)
        self.gmm_call = gmm.gmm_call(self.normalized_depth.normalized[region_name])

    def count_reads_diff_sites_and_variants(self):
        """Get read counts at differentiating sites and variant sites"""
        bam_handle = open_alignment_file(self.bamfile, self.reference)
        snp_db = self.call_parameters.snp_db
        self.gene1_read_count, self.gene2_read_count = get_supporting_reads(
            bam_handle, snp_db.dsnp1, snp_db.dsnp2, snp_db.nchr, snp_db.dindex
        )
        self.gene1_fraction = get_fraction(self.gene1_read_count, self.gene2_read_count)
        var_db = self.call_parameters.var_db
        self.var_alt, self.var_ref, var_alt_forward, var_alt_reverse = get_supporting_reads_single_region(
            bam_handle, var_db.dsnp1, var_db.nchr, var_db.dindex
        )
        var_homo_db = self.call_parameters.var_homo_db
        self.var_homo_alt, self.var_homo_ref = get_supporting_reads(
            bam_handle,
            var_homo_db.dsnp1,
            var_homo_db.dsnp2,
            var_homo_db.nchr,
            var_homo_db.dindex,
        )
        bam_handle.close()
        self.raw_count = OrderedDict()
        for i in range(len(self.var_list)):
            if i < len(self.var_alt):
                self.raw_count.setdefault(self.var_list[i], "%i,%i" % (self.var_alt[i], self.var_ref[i]))
            else:
                self.raw_count.setdefault(
                    self.var_list[i],
                    "%i,%i"
                    % (self.var_homo_alt[i - len(self.var_alt)], self.var_homo_ref[i - len(self.var_alt)]),
                )

    def call_cn_diff_sites_and_variants(self):
        """Call CN at differentiating sites and call variants"""
        # differentiating sites
        self.cn_call_snp = call_cn_snp(self.total_cn, self.gene1_read_count, self.gene2_read_count)
        self.raw_gene1_cn = [round(self.total_cn * a, 3) for a in self.gene1_fraction]
        # homology region
        cn_call_var_homo = call_cn_var_homo(
            self.total_cn, self.var_homo_alt, self.var_homo_ref, max_cn=2
            )
        # non-homology region
        cn_call_var = call_cn_var(2, self.var_alt, self.var_ref)
        # update variant names
        self.total_callset = get_called_variants(self.var_list, cn_call_var)
        called_var_homo = get_called_variants(self.var_list, cn_call_var_homo, len(cn_call_var))
        self.total_callset += called_var_homo




#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


import pysam


def parse_region_file(region_file):
    """Return the set of regions for counting from a bed file."""
    region_dic = {}
    with open(region_file) as read_region:
        for line in read_region:
            nchr, region_start, region_end, region_name, region_type, region_gc = (
                line.strip().split()
            )
            region_start = int(region_start)
            region_end = int(region_end)
            region = (nchr, region_start, region_end, region_name)
            region_dic.setdefault(region_type, []).append((region, region_gc))
    return region_dic


def parse_gmm_file(gmm_file):
    """Return the gmm parameters stored in input file."""
    dpar_tmp = {}
    with open(gmm_file) as read_gmm:
        for line in read_gmm:
            split_line = line.strip().split()
            dpar_tmp.setdefault(split_line[0], {})
            list_value = [a.split(":")[-1] for a in split_line[2:]]
            dpar_tmp[split_line[0]].setdefault(split_line[1], list_value)
    return dpar_tmp


def open_alignment_file(alignment_file, reference_fasta=None):
    if alignment_file.endswith("cram"):
        return pysam.AlignmentFile(
            alignment_file, "rc", reference_filename=reference_fasta
        )
    return pysam.AlignmentFile(alignment_file, "rb")



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from pprint import pprint
from collections import namedtuple, Counter
from .snp_count import get_supporting_reads
from .copy_number_call import call_cn_snp
from .haplotype import (
    get_haplotypes_from_bam,
    extract_hap,
    filter_hap,
    call_hap_cn,
    get_block_read_count,
)


class Phasing:
    """
    A class for phasing haplotypes across differentiating sites.
    """
    var_call = namedtuple(
        "var_call",
        "total_cn is_biallelic is_carrier depth_per_site deletion_bp_in_gene \
            full_haplotypes variants_allele1 variants_allele2 variant_details"
    )
    assembled_haplotypes = namedtuple(
        "assembled_haplotypes", "full_haplotypes hap_5p hap_3p hap_mid hap3p_merged"
    )
    var_call_per_haplotype = namedtuple(
        "var_call_per_haplotype", "haplotype variant_name block_result is_variant"
    )
    block_read_support = namedtuple(
        "block_read_support", "read_count flankings read_count_ordered likelihood"
    )

    CN_likelihood_threshold = 0.7
    CN_likelihood_threshold_loose = 0.45
    CN_likelihood_threshold_stringent = 0.95

    def __init__(self):
        self.variant_sites = []
        self.trusted_hap_threshold = []
        self.trusted_hap_switch_point = []
        self.variant_names = []
        self.variant_names_in_one = []
        self.flanking_site = {}
        self.bamfile = None
        self.base_db = None
        self.total_cn = None
        self.num_sites = None
        self.depth_calls = None
        self.reference = None
        self.haplotype_per_read = None
        self.gene1_read_count = []
        self.gene2_read_count = []
        self.deletion_bp_in_gene = None
        self.variants_details = []
        self.is_carrier = False
        self.is_biallelic = False
        self.variants_called_allele1 = []
        self.variants_called_allele2 = []
        self.depth_base_cn = None
        self.candidate_gene_hap = []

    def set_par(self, bamfile, base_db, total_cn, reference):
        """
        Establish parameter values.
        Parameters:
            bamfile: bam file
            base_db: differentiating site info
            total_cn: summed CN of paralogs
            reference: reference file
        """
        self.bamfile = bamfile
        self.base_db = base_db
        self.total_cn = total_cn
        self.reference = reference
        self.num_sites = len(base_db.dsnp1)
        self.gene1_read_count, self.gene2_read_count = get_supporting_reads(
            self.bamfile,
            self.base_db.dsnp1,
            self.base_db.dsnp2,
            self.base_db.nchr,
            self.base_db.dindex
        )
        self.depth_calls = call_cn_snp(
            self.total_cn,
            self.gene1_read_count,
            self.gene2_read_count
            )
        self.haplotype_per_read = get_haplotypes_from_bam(
            self.bamfile,
            self.base_db,
            range(self.num_sites),
            reference=self.reference
            )
        if total_cn < 4:
            self.deletion_bp_in_gene = False

    def carrier_from_assembled_haps(self, var_sites_haps):
        """
        Test if a sample is a carrier based on assembled haplotypes.
        Parameters:
            var_sites_haps (list of str): haplotypes only at variant sites
        Returns:
            bool
        """
        # Limit to total_cn=3
        # All haplotypes assembled. All unique.
        if len(var_sites_haps) == self.total_cn and self.total_cn == 3:
            gene_cn = [0] * len(self.variant_sites)
            for hap in var_sites_haps:
                for i in range(len(hap)):
                    if hap[i] == "1":
                        gene_cn[i] += 1
            # CN(gene) is 1 in fully assembled haplotypes.
            if 1 in gene_cn:
                return True
        return False

    def get_variants(self, hap):
        """
        Get variant names given haplotype
        Parameters:
            hap (str): an assembed haplotype
        Returns:
            variants (list of str): names of variants carried by the haplotype
        """
        variants = []
        for i in range(len(hap)):
            if hap[i] == "2" and i in self.variant_sites:
                variant_index = self.variant_sites.index(i)
                variants.append(self.variant_names[variant_index])
        return variants

    def assess_assembled_haps(self, full_haplotypes):
        """
        Check if haplotypes have been completely assembled.
        Parameters:
            full_haplotypes (list of str): assembled haplotypes
        Returns:
            fully_assembled (bool)
            var_sites_haps (list of str): haplotypes only at variant sites
        """
        var_sites_haps = []
        fully_assembled = True
        for hap in full_haplotypes:
            if 'x' in hap:
                fully_assembled = False
            else:
                sites_of_interest = ''.join(
                    [hap[i] for i in self.variant_sites]
                    )
                var_sites_haps.append(sites_of_interest)
        return fully_assembled, var_sites_haps

    def filter_haps_get_var_sites(self, full_haplotypes, trust_hap=False):
        """
        From haplotypes, output variant sites (switching points)
        Parameters:
            full_haplotypes (list of str): assembled haplotypes
            trust_hap (bool): whether to trust trusted haplotypes
        Returns:
            var_index_in_haps (dict of str:list): index of candidate
            variant sites on each haplotype
        """
        var_index_in_haps = {}
        for hap in full_haplotypes:
            count_2 = hap.count("2")
            count_1 = self.depth_calls.count(1)
            count_0 = self.depth_calls.count(0)
            count_none = self.depth_calls.count(None)
            # this is the criteria for selecting candidate haplotypes,
            # comparing haplotypes against the depth calls
            if (
                count_2 <= count_1 + count_0 + count_none + 1 or
                (trust_hap and hap in self.trusted_hap)
            ):
                for var_index in range(self.num_sites):
                    if hap[var_index] == "2":
                        # find the first variant site past the switching point
                        if (var_index > 0 and hap[var_index - 1] == "1"):
                            n = var_index
                            while n < self.num_sites:
                                if n in self.variant_sites and hap[n] == "2":
                                    var_index_in_haps.setdefault(hap, [])
                                    if n not in var_index_in_haps[hap]:
                                        var_index_in_haps[hap].append(n)
                                    break
                                n += 1
                        if (
                            var_index + 1 < self.num_sites and
                            hap[var_index + 1] == "1"
                            ):
                            n = var_index
                            while n >= 0:
                                if n in self.variant_sites and hap[n] == "2":
                                    var_index_in_haps.setdefault(hap, [])
                                    if n not in var_index_in_haps[hap]:
                                        var_index_in_haps[hap].append(n)
                                    break
                                n -= 1
        return var_index_in_haps

    def get_depth_for_block(self, flanking_sites_per_var, var_index_in_flanking):
        """
        from single switching point, check if read counts
        support that there is only one copy of WT gene
        Parameters:
            flanking_sites_per_var (list of int): sites flanking a variant to check depth
            var_index_in_flanking (int): index of variant site in haplotypes
        Returns:
            named tuple (block_read_support) summarizing depth info at this site
        """
        switch_site_haps_count = extract_hap(
                self.haplotype_per_read, range(flanking_sites_per_var[0], flanking_sites_per_var[1] + 1)
            )
        switch_site_haps_count = filter_hap(switch_site_haps_count)
        if (
            self.total_cn == len(switch_site_haps_count) and
            min(switch_site_haps_count.values()) >= 2
        ):
            return(
                self.block_read_support(switch_site_haps_count, flanking_sites_per_var, None, 1)
                )
        if self.total_cn >= 2:
            block_read_count = get_block_read_count(switch_site_haps_count, var_index_in_flanking)
            if block_read_count is not None:
                if block_read_count == [0]:
                    return(
                        self.block_read_support(
                            switch_site_haps_count,
                            flanking_sites_per_var,
                            block_read_count,
                            1
                            )
                        )
                cn_call = call_hap_cn(block_read_count, self.total_cn)
                return(
                    self.block_read_support(
                        switch_site_haps_count,
                        flanking_sites_per_var,
                        block_read_count,
                        round(cn_call, 4)
                        )
                    )
        return(self.block_read_support(switch_site_haps_count, flanking_sites_per_var, None, None))

    def get_depth_for_blocks(self, good_var_indices, cn_threshold):
        """
        from a set of switching points, check if read counts
        support that there is only one copy of WT gene
        Parameters:
            good_var_indices (list of int): indices of variant sites on a haplotype
            cn_threshold (float): likelihood cutoff
        Returns:
            found_variant_in_short_hap (bool): whether a variant is found based on depth
            blocks (list of get_depth_for_block named tuple): summary of depth info
        """
        found_variant_in_short_hap = False
        blocks = []
        for var_index in good_var_indices:
            for flanking_sites_per_var in self.flanking_site[var_index]:
                var_index_in_flanking = None
                sites_range = list(range(flanking_sites_per_var[0], flanking_sites_per_var[1] + 1))
                if var_index in sites_range:
                    var_index_in_flanking = sites_range.index(var_index)
                block = self.get_depth_for_block(flanking_sites_per_var, var_index_in_flanking)
                blocks.append(block)
                if block.likelihood is not None and block.likelihood > cn_threshold:
                    found_variant_in_short_hap = True
        return found_variant_in_short_hap, blocks

    def compare_haplotype_against_depth(self, hap):
        """
        In rare cases there is more than one candidate haplotype.
        Pick the one that best explains the depth values
        Parameters:
            hap (list of str): a haplotype
        Returns:
            (float) ratio of matches compared to depth paterns
        """
        assert len(hap) == len(self.depth_calls)
        nmatch = 0
        nsites = len(hap)
        for i in range(nsites):
            if self.depth_calls[i] is not None:
                if (
                    (hap[i] == '2' and self.depth_calls[i] == 1) or
                    (hap[i] == '1' and self.depth_calls[i] >= 2)
                ):
                    nmatch += 1
        return nmatch/nsites

    def check_complete_loss(self, var_index_in_haps):
        """completely lose one gene, or one long conversion"""
        if var_index_in_haps == {}:
            gene1_total = sum(self.gene1_read_count)
            gene2_total = sum(self.gene2_read_count)
            combined_cn_call = call_cn_snp(
                self.total_cn,
                [gene1_total],
                [gene2_total],
                0.8
                )
            if combined_cn_call[0] is not None and combined_cn_call[0] == 0 and self.total_cn < 4:
                self.deletion_bp_in_gene = True
            if combined_cn_call[0] is not None and combined_cn_call[0] == 1:
                count_1 = self.depth_calls.count(1)
                count_none = self.depth_calls.count(None)
                if (
                    count_1 + count_none >= len(self.depth_calls) - 1 and
                    count_1 > count_none
                ):
                    self.is_carrier = True
                    self.variants_details.append(
                        self.var_call_per_haplotype(
                            None,
                            self.variant_names_in_one,
                            None,
                            True
                            )
                        )
                    if self.total_cn < 4:
                        self.deletion_bp_in_gene = True

    def assess_compound_het(self, full_haplotypes):
        """assess compound het"""
        fully_assembled, var_sites_haps = self.assess_assembled_haps(full_haplotypes)
        if fully_assembled is True:
            # Among all haplotypes,
            # there is not a copy that are WT in all positions
            if "1"*len(self.variant_sites) not in var_sites_haps:
                return True
        return False

    def assess_homozygous(self):
        """assess if there are homozygous sites"""
        homozygous_depth = False
        for i in self.variant_sites:
            if self.depth_calls[i] is not None and self.depth_calls[i] == 0:
                homozygous_depth = True
                self.variants_called_allele1.append(self.variant_names[self.variant_sites.index(i)])
                self.variants_called_allele2.append(self.variant_names[self.variant_sites.index(i)])
        return homozygous_depth

    def pick_best_variant(self, compound_het):
        """
        In rare scenarios where sample is not compound het but
        there are more than one variant haplotypes,
        pick one that best fits depth calls
        Parameters:
            compound_het (bool): whether sample is compound het
        """
        unique_vars = []
        unique_haps = []
        for var in self.variants_details:
            if var.is_variant:
                if var.variant_name not in unique_vars:
                    unique_vars.append(var.variant_name)
                if var.haplotype not in unique_haps:
                    unique_haps.append(var.haplotype)
        if len(unique_haps) > 1 and compound_het is False:
            hap_score = []
            var_list = []
            for var in self.variants_details:
                if var.is_variant:
                    var_list.append(var)
                    hap_score.append(self.compare_haplotype_against_depth(var.haplotype))
            best_vars = []
            max_score = max(hap_score)
            for i,score in enumerate(hap_score):
                if score == max_score:
                    best_vars.append(var_list[i].variant_name)
            best_vars = sorted(best_vars, key=lambda kv: len(kv))
            best_var = best_vars[0]
            variants_called_pick_one = []
            for var in self.variants_details:
                if var.variant_name == best_var:
                    variants_called_pick_one.append(var)
                    unique_vars = [var.variant_name]
                else:
                    variants_called_pick_one.append(var._replace(is_variant=False))
            self.variants_details = variants_called_pick_one
        unique_vars = sorted(unique_vars, key=lambda kv: len(kv))
        return unique_vars

    def get_candidate_gene_hap(self, full_haplotypes):
        """
        Get haplotypes that are (roughly) likely from the gene based
        on the number of gene bases
        """
        for haplotype in full_haplotypes:
            if haplotype.count("1") > len(haplotype)/2:
                self.candidate_gene_hap.append(haplotype)

    def get_base_cn(self, start, end):
        """
        Get (roughly) the number of gene haplotypes from depth calls
        Updates depth_base_cn
        Parameters:
            start (int): start position to count depth calls
            end (int): end position to count depth calls
        """
        clean_depth = []
        for i, site in enumerate(self.depth_calls):
            if start <= i <= end:
                if site is not None:
                    clean_depth.append(site)
        clean_depth_counter = sorted(
            Counter(clean_depth).items(), key=lambda kv: kv[1], reverse=True
        )
        if (
            clean_depth_counter[0] is not None and
            clean_depth_counter[0][1] >= len(self.depth_calls)/2
        ):
            self.depth_base_cn = clean_depth_counter[0][0]

    def check_deletion_bp_in_gene(self, var_index_in_haps):
        """
        In case of a deletion,
        check whether the deletion breakpoint is within the gene
        """
        return False

    def assess_hap(self, hap, var_index_in_haps, full_haplotypes):
        """
        Assess whether a haplotype is a variant haplotype that makes
        the sample a carrier i.e. having only one copy of the WT gene
        """
        return False

    def assemble_haplotypes(self, debug=False):
        """Assemble haplotypes that consist of differentiating sites"""
        return self.assembled_haplotypes([], None, None, None, None)

    def call_variants_from_haplotype(self, debug=False):
        """
        Call recombinant variants by phasing through differentiating sites
        """
        if debug is True:
            pprint(self.haplotype_per_read)

        # assemble haplotypes
        haps_assembled = self.assemble_haplotypes(debug=debug)
        full_haplotypes = haps_assembled.full_haplotypes
        self.get_candidate_gene_hap(full_haplotypes)

        homozygous_depth = self.assess_homozygous()
        compound_het = self.assess_compound_het(full_haplotypes)

        var_index_in_haps = self.filter_haps_get_var_sites(full_haplotypes)
        self.check_deletion_bp_in_gene(var_index_in_haps)
        # assess each haplotype to see if they makes the sample a carrier,
        # i.e. having one copy of the WT gene
        for hap in var_index_in_haps:
            self.assess_hap(hap, var_index_in_haps, full_haplotypes)

        # completely lose one gene, or one long conversion across all sites
        self.check_complete_loss(var_index_in_haps)

        unique_vars = self.pick_best_variant(compound_het)
        self.is_biallelic = homozygous_depth or (compound_het and len(unique_vars) >= 2)
        if self.is_biallelic:
            self.is_carrier = False
        if unique_vars != []:
            self.variants_called_allele1 += unique_vars[0]
            if len(unique_vars) > 1:
                self.variants_called_allele2 += unique_vars[1]

        return self.var_call(
            self.total_cn,
            self.is_biallelic,
            self.is_carrier,
            '_'.join([str(a) for a in self.depth_calls]),
            self.deletion_bp_in_gene,
            haps_assembled._asdict(),
            list(set(self.variants_called_allele1)),
            list(set(self.variants_called_allele2)),
            [a._asdict() for a in self.variants_details]
            )



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from scipy.stats import poisson
import re
import pysam
from collections import namedtuple
from .snp_count import reverse_complement

ERROR_RATE = 0.01
haplotype_groups = namedtuple(
    "haplotype_groups",
    "unique nonunique"
)
clip_start_pat = re.compile(r"^(\d+)S")
clip_end_pat = re.compile(r"(\d+)S$")


def get_NM(ltag):
    """Get NM tag for a read alignment"""
    for tag in ltag:
        if tag[0] == "NM":
            return tag[1]
    return None


def get_mismatches_from_softclip(read_alignment, ref, min_base_quality=13):
    """
    Get number of mismatches in soft-clipped sequences, considering base quality
    Sometimes a read is soft-clipped due to low base quality, so the size of the clip
    is not necessarily equal to the length of a mismatch.
    We want to keep such reads so the unclipped part is informative.
    """
    mismatch = 0
    if ref is None:
        return mismatch
    ref_pos = read_alignment.reference_start
    ref_end = read_alignment.reference_end
    cigar = read_alignment.cigarstring
    read_len = len(read_alignment.query_sequence)
    left_clip = sum([int(a) for a in clip_start_pat.findall(cigar)])
    right_clip = sum([int(a) for a in clip_end_pat.findall(cigar)])
    if left_clip > 0:
        for i in range(left_clip):
            if read_alignment.query_qualities[left_clip - (i+1)] >= min_base_quality:
                pos = ref_pos - (i+1)
                ref_base = ref.fetch(read_alignment.reference_name, pos, pos+1).upper()
                read_base = read_alignment.query_sequence[left_clip - (i+1)].upper()
                if ref_base != read_base:
                    mismatch += 1
    if right_clip > 0:
        for i in range(right_clip):
            if read_alignment.query_qualities[read_len - (right_clip - i)] >= min_base_quality:
                pos = ref_end + i
                ref_base = ref.fetch(read_alignment.reference_name, pos, pos+1).upper()
                read_base = read_alignment.query_sequence[read_len - (right_clip - i)].upper()
                if ref_base != read_base:
                    mismatch += 1
    return mismatch


def passing_read(pileupread, ref=None, long_read=False):
    """Return whether a read passes filter."""
    nm = get_NM(pileupread.alignment.tags)
    start_clip = sum([int(a) for a in clip_start_pat.findall(pileupread.alignment.cigarstring)])
    end_clip = sum([int(a) for a in clip_end_pat.findall(pileupread.alignment.cigarstring)])
    clip = start_clip + end_clip
    clip_mismatch = 0
    if ref is not None and clip > 15:
        clip_mismatch = get_mismatches_from_softclip(pileupread.alignment, ref)
    return (
        not pileupread.is_del
        and not pileupread.is_refskip
        and pileupread.alignment.is_secondary == 0
        and pileupread.alignment.is_supplementary == 0
        and pileupread.alignment.is_duplicate == 0
        and (nm is None or (nm is not None and nm <= 6) or long_read)
        and (
            (ref is None and clip <= 15)
            or (ref is not None and clip_mismatch <= 10)
        )
    )


def get_haplotypes_from_bam(bamfile_handle, base_db, target_positions, rc=None, reference=None):
    """
    Get haplotypes for each read at a set of target positions
    Parameters:
        bamfile_handle
        base_db (named tuple): differentiating site info
        target_positions (list of int): positions of interest
        rc (bool): reverse complement if any gene is on the reverse strand
        reference (str): reference fasta
    Returns:
        haplotype_per_read (dict of str : str): bases of each read at differentiating sites
    """
    dread = {}
    dread = get_bases_per_read(bamfile_handle, base_db, target_positions, rc=rc, reference=reference)
    base1, base2 = get_base1_base2(base_db, target_positions)
    haplotype_per_read = get_haplotype_per_read(dread, base1, base2)
    return haplotype_per_read


def get_haplotypes_from_bam_single_region(bamfile_handle, base_db, target_positions, rc=None, reference=None):
    """
    Get haplotypes for each read at a set of target positions, from only the gene region
    Parameters:
        bamfile_handle
        base_db (named tuple): differentiating site info
        target_positions (list of int): positions of interest
        rc (bool): reverse complement if any gene is on the reverse strand
        reference (str): reference fasta
    Returns:
        haplotype_per_read (dict of str : str): bases of each read at differentiating sites
    """
    dread = {}
    dread = get_bases_per_read(
        bamfile_handle, base_db, target_positions, region=0, min_mapq=10, rc=rc, reference=reference
    )
    base1, base2 = get_base1_base2(base_db, target_positions)
    haplotype_per_read = get_haplotype_per_read(dread, base1, base2)
    return haplotype_per_read


def get_bases_per_read(
    bamfile_handle, base_db, target_positions, region=None, min_mapq=0, rc=None, reference=None
):
    """
    Go through reads and get bases at sites of interest
    Paramters:
        bamfile_handle
        base_db (named tuple): differentiating site info
        target_positions (list of int): positions of interest
        region (int): analyze both regions (gene and pseudogene) or just one.
        min_mapq (int): mapq cutoff
        rc (bool): reverse complement if any gene is on the reverse strand
        reference (str): reference fasta
    Returns:
        dread (dict of str:list): bases at each site on individual reads,
        before converting to 1s and 2s
    """
    ref = None
    if reference is not None:
        ref = pysam.Fastafile(reference)
    dread = {}
    nchr = base_db.nchr
    dindex = base_db.dindex
    dsnps = [base_db.dsnp1, base_db.dsnp2]
    if region is not None:
        if region == 0:
            dsnps = [base_db.dsnp1]
        elif region == 1:
            dsnps = [base_db.dsnp2]
    for i in range(len(dsnps)):
        dsnp = dsnps[i]
        for snp_position_ori in dsnp:
            dsnp_index = dindex[snp_position_ori]
            snp_position = int(snp_position_ori.split("_")[0])
            if dsnp_index in target_positions:
                for pileupcolumn in bamfile_handle.pileup(
                    nchr,
                    snp_position - 1,
                    snp_position + 1,
                    truncate=True,
                    stepper="nofilter",
                    ignore_overlaps=False,
                    ignore_orphan=False,
                    min_base_quality=13,
                ):
                    site_position = pileupcolumn.pos + 1
                    if site_position == snp_position:
                        reg1_allele, reg2_allele = dsnp[snp_position_ori].split("_")
                        for read in pileupcolumn.pileups:
                            if (
                                passing_read(read, ref=ref)
                                and read.alignment.mapping_quality >= min_mapq
                            ):
                                read_name = read.alignment.query_name
                                read_seq = read.alignment.query_sequence
                                start_pos = read.query_position
                                end_pos = start_pos + min(
                                    len(reg1_allele), len(reg2_allele)
                                )
                                if end_pos < len(read_seq):
                                    hap = read_seq[start_pos:end_pos]
                                    if i == rc:
                                        hap = reverse_complement(hap)
                                    if read_name not in dread:
                                        dread.setdefault(read_name, {})
                                        for pos in target_positions:
                                            dread[read_name].setdefault(pos, None)
                                    if dread[read_name][dsnp_index] not in [None, hap]:
                                        dread[read_name][dsnp_index] = None
                                    dread[read_name][dsnp_index] = hap
    return dread


def get_haplotype_per_read(dread, base1, base2):
    """
    Translate bases into haplotypes, labeled as 1 or 2
    Parameters:
        dread (dict of str:list): bases at each site on individual reads
        base1 (list of str): gene1 bases
        base2 (list of str): gene2 bases
    Returns:
        haplotype_per_read (dict of str : str): bases of each read at differentiating sites
    """
    dread_hap = {}
    for read in dread:
        read_bases = dread[read]
        assert len(read_bases) == len(base1) == len(base2)
        pos_list = ["x"] * len(read_bases)
        for i, pos in enumerate(read_bases):
            base = read_bases[pos]
            if base is not None:
                for allele in base1[i].split(","):
                    if base == allele.upper():
                        pos_list[i] = "1"
                for allele in base2[i].split(","):
                    if base == allele.upper():
                        pos_list[i] = "2"
        dread_hap.setdefault(read, "".join(pos_list))
    return dread_hap


def get_base1_base2(base_db, target_positions):
    """
    Get expected bases corresponding to different alleles/paralogs
    at target positions
    Parameters:
        base_db (named tuple): differentiating site info
        target_positions (list of int): positions of interest
    Returns:
        list of gene1 bases, list of gene2 bases
    """
    base1 = [None] * len(base_db.dsnp1)
    base2 = [None] * len(base_db.dsnp1)
    dindex = base_db.dindex
    for pos in base_db.dsnp1:
        dsnp_index = dindex[pos]
        if dsnp_index in target_positions:
            index = int(pos.split("_")[1])
            allele1, allele2 = base_db.dsnp1[pos].split("_")
            base1[index] = allele1
            base2[index] = allele2
    return [base1[i] for i in target_positions], [base2[i] for i in target_positions]


def extract_hap(haplotype_per_read, positions_to_extract):
    """
    Extract haplotypes at given positions
    Paramters:
        haplotype_per_read (dict of str:str): bases of each read at differentiating sites
        positions_to_extract (list of int): positions of interest
    Returns:
        hap_count (dict of str:list): haplotypes and a list of supporting reads (as 1)
    """
    hap_count = {}
    for read_name in haplotype_per_read:
        hap = haplotype_per_read[read_name]
        hap_base = [hap[pos] for pos in positions_to_extract]
        if "x" not in hap_base:
            hap_count.setdefault("".join(hap_base), []).append(1)
    return hap_count


def group_haps(haplotype_per_read, hap_list):
    """
    Assign reads to partial haplotype blocks so haplotype blocks can be extended
    Parameters:
        haplotype_per_read (dict of str:str): bases of each read at differentiating sites
        hap_list (list of str): haplotypes to be extended
    Returns:
        (named tuple of dict of str:list): reads grouped with each haplotype,
        for uniquely and nonuniquely associated reads
    """
    matching_haplotype_groups = {}
    matching_haplotype_groups_nonunique = {}
    for read_name in haplotype_per_read:
        read_hap = haplotype_per_read[read_name]
        matching_haplotypes = []
        for haplotype_to_extend in hap_list:
            match, mismatch, extend = compare_two_haps(haplotype_to_extend, read_hap)
            if mismatch == 0 and match > 0 and extend >= 0:
                matching_haplotypes.append(haplotype_to_extend)
                matching_haplotype_groups_nonunique.setdefault(haplotype_to_extend, []).append(read_hap)
        # unique match
        if len(matching_haplotypes) == 1:
            matching_haplotype_groups.setdefault(matching_haplotypes[0], []).append(read_hap)
    return haplotype_groups(matching_haplotype_groups, matching_haplotype_groups_nonunique)


def extend_hap_5p(matching_haplotype_groups, min_read=None):
    """
    Extend a haplotype on 5p based on reads
        Parameters:
        matching_haplotype_groups (named tuple of dict of str:list):
        reads grouped with each haplotype to be extended, for uniquely
        and nonuniquely associated reads
        min_read (int): minimum read count for extension
    Returns:
        new_hap_list (list of str): left extended haplotypes
    """
    new_hap_list = []
    for haplotype_to_extend in matching_haplotype_groups.unique:
        if "x" in haplotype_to_extend:
            new_hap = haplotype_to_extend
            haps = matching_haplotype_groups.unique[haplotype_to_extend]
            pos_x = 0
            # find the first non-"x" position to start extending
            while pos_x < len(new_hap):
                if new_hap[pos_x] != "x":
                    break
                pos_x += 1
            if pos_x == 0:
                new_hap_list.append(haplotype_to_extend)
            else:
                extended = False
                pos = pos_x
                i = 0
                # going to the left site by site and
                # get consensus of the base from matching read haplotypes
                while pos > 0:
                    pos = pos - 1
                    i += 1
                    bases_at_site = [a[pos] for a in haps if a[pos] != "x"]
                    if bases_at_site != []:
                        uniq_bases = []
                        for uniq_base in list(set(bases_at_site)):
                            min_read_site = min_read
                            if min_read is None:
                                min_read_site = get_read_count_threshold(bases_at_site)
                            if bases_at_site.count(uniq_base) >= min_read_site:
                                uniq_bases.append(uniq_base)
                        for uniq_base in uniq_bases:
                            new_extended_haplotype = "x" * pos + uniq_base + new_hap[(pos + 1) :]
                            new_hap_list.append(new_extended_haplotype)
                            extended = True
                        break
                if extended is False and matching_haplotype_groups.nonunique:
                    pos = pos_x - 1
                    next_pos_nonuniques = [
                        a[pos] for a in matching_haplotype_groups.nonunique[haplotype_to_extend] if a[pos] != "x"
                        ]
                    if len(set(next_pos_nonuniques)) == 1 and len(next_pos_nonuniques) > 4:
                        new_extended_haplotype = "x" * pos + next_pos_nonuniques[0] + new_hap[(pos + 1) :]
                        new_hap_list.append(new_extended_haplotype)
                        extended = True
                # failed to extend
                if extended is False:
                    new_hap_list.append(haplotype_to_extend)
        else:
            new_hap_list.append(haplotype_to_extend)
    return new_hap_list


def extend_hap_3p(matching_haplotype_groups, min_read=None):
    """
    Extend a haplotype on 3p based on reads
    Parameters:
        matching_haplotype_groups (named tuple of dict of str:list):
        reads grouped with each haplotype to be extended, for uniquely
        and nonuniquely associated reads
        min_read (int): minimum read count for extension
    Returns:
        new_hap_list (list of str): right extended haplotypes
    """
    new_hap_list = []
    for haplotype_to_extend in matching_haplotype_groups.unique:
        if "x" in haplotype_to_extend:
            new_hap = haplotype_to_extend
            hap_len = len(new_hap)
            haps = matching_haplotype_groups.unique[haplotype_to_extend]
            pos_x = hap_len - 1
            # find the first non-"x" position from the 3p end to start extending
            while pos_x >= 0:
                if new_hap[pos_x] != "x":
                    break
                pos_x -= 1
            if pos_x == hap_len - 1:
                new_hap_list.append(haplotype_to_extend)
            else:
                extended = False
                pos = pos_x
                i = 0
                while pos < hap_len - 1:
                    pos = pos + 1
                    i += 1
                    bases_at_site = [a[pos] for a in haps if a[pos] != "x"]
                    if bases_at_site != []:
                        uniq_bases = []
                        for uniq_base in list(set(bases_at_site)):
                            min_read_site = min_read
                            if min_read is None:
                                min_read_site = get_read_count_threshold(bases_at_site)
                            if bases_at_site.count(uniq_base) >= min_read_site:
                                uniq_bases.append(uniq_base)
                        for uniq_base in uniq_bases:
                            new_extended_haplotype = new_hap[:pos] + uniq_base + "x" * (hap_len - 1 - pos)
                            new_hap_list.append(new_extended_haplotype)
                            extended = True
                        break
                if extended is False and matching_haplotype_groups.nonunique:
                    pos = pos_x + 1
                    next_pos_nonuniques = [
                        a[pos] for a in matching_haplotype_groups.nonunique[haplotype_to_extend] if a[pos] != "x"
                        ]
                    if len(set(next_pos_nonuniques)) == 1 and len(next_pos_nonuniques) > 4:
                        new_extended_haplotype = new_hap[:pos] + next_pos_nonuniques[0] + "x" * (hap_len - 1 - pos)
                        new_hap_list.append(new_extended_haplotype)
                        extended = True
                if extended is False:
                    new_hap_list.append(haplotype_to_extend)
        else:
            new_hap_list.append(haplotype_to_extend)
    return new_hap_list


def get_block_read_count(haps, var_index_in_flanking):
    """
    Return supporting reads for different haplotypes.
    Parameters:
        haps (dict of str: int): read counts of haplotypes
        var_index_in_flanking (int): index of variant site in haplotypes
    Returns:
        new_lcount (list of int): reordered read counts. first one should be WT gene
    """
    index_gene = None
    index_pseudo = None
    index_var = None
    wt_reads = 0
    lcount = []
    for i, hap in enumerate(haps):
        count = haps[hap]
        lcount.append(count)
        if hap[var_index_in_flanking] == '1':
            wt_reads += count
        if hap.count("1") == len(hap):
            index_gene = i
        elif hap.count("2") == len(hap):
            index_pseudo = i
        else:
            index_var = i
    # if no read carries WT gene base at the variant site
    if wt_reads == 0 and sum(lcount) > 10:
        return [0]
    new_lcount = None
    if index_gene is None:
        return None
    else:
        new_lcount = [lcount[index_gene]]
        if len(haps) == 2:
            new_lcount.append(lcount[1-index_gene])
        elif len(haps) == 3:
            if index_pseudo is not None:
                new_lcount.append(lcount[index_var])
                new_lcount.append(lcount[index_pseudo])
            else:
                for i in range(len(haps)):
                    if i != index_gene:
                        new_lcount.append(lcount[i])
        elif len(haps) == 4:
            for i in range(len(haps)):
                if i != index_gene:
                    new_lcount.append(lcount[i])
        else:
            return None
    return new_lcount


def get_possible_hap_cns(full_cn, haplotype_count):
    """
    Given the total copy number (3 or 4) and the number of haplotypes,
    Output all possible assignments of copy number per haplotype
    Parameters:
        full_cn (int): total cn of paralogs
        haplotype_count (int): number of haplotypes
    Returns:
        possible_hap_cns(list of tuples): each item is a tuple of
        size equal to haplotype_count, containing copy numbers of
        each haplotype
    """
    possible_hap_cns = set()
    if haplotype_count == 3:
        for i in range(1, full_cn - haplotype_count + 2):
            for j in range(1, full_cn - haplotype_count + 2):
                for k in range(1, full_cn - haplotype_count + 2):
                    if i+j+k == full_cn:
                        possible_hap_cns.add((i, j, k))
    if haplotype_count == 4:
        for i in range(1, full_cn - haplotype_count + 2):
            for j in range(1, full_cn - haplotype_count + 2):
                for k in range(1, full_cn - haplotype_count + 2):
                    for l in range(1, full_cn - haplotype_count + 2):
                        if i+j+k+l == full_cn:
                            possible_hap_cns.add((i, j, k, l))
    return possible_hap_cns


def call_hap_cn(lcount, full_cn):
    """
    Likelihood of having only one copy of WT gene, based on depth of haplotypes
    Parameters:
        lcount (list of int): read counts of haplotypes. first one should be WT gene
        full_cn (int): total cn of paralogs
    Returns:
        likelihood of the first haplotype having a cn of 1
    """
    nsum = sum(lcount)
    if full_cn == len(lcount):
        return 1
    if (
        nsum == 0 or
        full_cn < len(lcount) or
        0 in lcount or
        len(lcount) > 4 or
        full_cn >= 10
    ):
        return 0

    gene_count = lcount[0]
    # below are possible scenarios of copie numbers of WT, variant, pseudo
    if len(lcount) == 2:
        # 2 haplotypes, full_cn > 2
        prob = []
        for i in range(full_cn):
            if i > 0:
                prob_value = 1
                depthexpected = nsum * i / full_cn
                prob_value *= poisson.pmf(gene_count, depthexpected)
                depthexpected = nsum * (full_cn - i) / full_cn
                prob_value *= poisson.pmf(lcount[1], depthexpected)
                prob.append(prob_value)
        sum_prob = sum(prob)
        if sum_prob == 0:
            return 0
        post_prob = [a/sum_prob for a in prob]
        return post_prob[0]

    possible_hap_cns = get_possible_hap_cns(full_cn, len(lcount))
    prob = {}
    if len(lcount) == 3:
        # 3 haplotypes, full_cn > 3
        for possible_hap_cn in possible_hap_cns:
            prob_value = 1
            depthexpected = nsum * possible_hap_cn[0] / full_cn
            prob_value *= poisson.pmf(lcount[0], depthexpected)
            depthexpected = nsum * possible_hap_cn[1] / full_cn
            prob_value *= poisson.pmf(lcount[1], depthexpected)
            depthexpected = nsum * possible_hap_cn[2] / full_cn
            prob_value *= poisson.pmf(lcount[2], depthexpected)
            prob.setdefault(possible_hap_cn, prob_value)
    if len(lcount) == 4:
        # 4 haplotypes, full_cn > 4
        for possible_hap_cn in possible_hap_cns:
            prob_value = 1
            depthexpected = nsum * possible_hap_cn[0] / full_cn
            prob_value *= poisson.pmf(lcount[0], depthexpected)
            depthexpected = nsum * possible_hap_cn[1] / full_cn
            prob_value *= poisson.pmf(lcount[1], depthexpected)
            depthexpected = nsum * possible_hap_cn[2] / full_cn
            prob_value *= poisson.pmf(lcount[2], depthexpected)
            depthexpected = nsum * possible_hap_cn[3] / full_cn
            prob_value *= poisson.pmf(lcount[3], depthexpected)
            prob.setdefault(possible_hap_cn, prob_value)
    #print(lcount, full_cn, prob)
    sum_prob = sum(prob.values())
    if sum_prob == 0:
        return 0
    post_prob = []
    for possible_hap_cn in prob:
        if possible_hap_cn[0] == 1:
            post_prob.append(prob[possible_hap_cn]/sum_prob)
    if post_prob != []:
        return max(post_prob)
    return 0


def get_read_count_threshold(bases_at_site):
    """
    Get read count threshold depending on counts
    Parameters:
        bases_at_site (list of str): list of bases seen at a site
    Returns:
        min_count (int): minimum supporting reads to filter on
    """
    base_count = []
    for uniq_base in list(set(bases_at_site)):
        base_count.append(bases_at_site.count(uniq_base))
    min_count = 2
    if max(base_count) <= 4:
        min_count = 1
    return min_count


def join_subblocks(haplotype_per_read, lhap1, lhap2, pos1, pos2):
    """
    Merge two partially assembled haplotypes
    Parameters:
        haplotype_per_read (dict of str : str): bases of each read at differentiating sites
        lhap1 (list of str): haplotypes to be merged
        lhap2 (list of str): haplotypes to be merged
        pos1: looks for read support at one side before this position
        pos2: looks for read support at other side after this position
    Returns:
        full_hap (list of str): merged haplotypes
        dcount (dict of str : list): reads supporting each of the merged haplotypes
    """
    full_hap = []
    for sequence1 in lhap1:
        for sequence2 in lhap2:
            # Check if the pos1-pos2 region matches between the two sequences
            consensus = get_seq_consensus(sequence1, sequence2)
            if consensus is not False and consensus not in full_hap:
                    full_hap.append(consensus)
    dcount = {}
    for merged_hap in full_hap:
        for read_name in haplotype_per_read:
            read_hap = haplotype_per_read[read_name]
            match1, mismatch1, extend1 = compare_two_haps(read_hap[:pos1], merged_hap[:pos1])
            match2, mismatch2, extend2 = compare_two_haps(read_hap[pos2:], merged_hap[pos2:])
            if match1 > 0 and match2 > 0 and mismatch1 == 0 and mismatch2 == 0:
                dcount.setdefault(merged_hap, []).append(read_name)
    return full_hap, dcount


def filter_hap(haplotype_read_count, min_read=2):
    """
    Get haplotypes' supporting read count, filter by minimum read count
    Parameters:
        haplotype_read_count (dic of str : list): reads supporting each haplotype
        min_read (int): minimum supporting read count
    Returns:
        haplotype_read_count_filtered (dic of str : int): number of supporting reads
    """
    haplotype_read_count_filtered = {}
    for hap_item in haplotype_read_count:
        hap_count = sum(haplotype_read_count[hap_item])
        if hap_count >= min_read:
            haplotype_read_count_filtered.setdefault(hap_item, hap_count)
    return haplotype_read_count_filtered


def compare_two_haps(hap1, hap2):
    """
    Compare two haplotypes, and calculate number of bases for match/mismatch/extend
    Parameters:
        hap1 (str): sequence of haplotype 1
        hap2 (str): sequence of haplotype 2
    Returns:
        match (int): number of matching bases
        mismatch (int): number of mismatching bases
        extend (int): number of extending bases
    """
    assert len(hap1) == len(hap2)
    match = 0
    mismatch = 0
    extend = 0
    for i in range(len(hap1)):
        base1 = hap1[i]
        base2 = hap2[i]
        if "x" not in [base1, base2]:
            if base1 == base2:
                match += 1
            else:
                mismatch += 1
        if base1 == "x" and base2 != "x":
            extend += 1
    return match, mismatch, extend


def get_seq_consensus(hap1, hap2):
    """
    compare two haplotype blocks and produce consensus
    Parameters:
        hap1 (str): sequence of haplotype 1
        hap2 (str): sequence of haplotype 2
    Returns:
        consensus (str): consensus sequence
    """
    assert len(hap1) == len(hap2)
    if hap1.count("x") == len(hap1) or hap2.count("x") == len(hap2):
        return False
    consensus = ""
    match = 0
    for i in range(len(hap1)):
        base1 = hap1[i]
        base2 = hap2[i]
        if "x" not in [base1, base2]:
            if base1 != base2:
                return False
            match += 1
            consensus += base1
        elif base1 != "x":
            consensus += base1
        elif base2 != "x":
            consensus += base2
        else:
            consensus += "x"
    if match == 0:
        return False
    return consensus



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from scipy.stats import poisson

POSTERIOR_CUTOFF_STRINGENT = 0.9
ERROR_RATE = 0.01


def call_reg1_cn(full_cn, count_reg1, count_reg2, min_read=0):
    """
    Return the reg1 copy number call at each site based on Poisson likelihood,
    with a minimum read support cutoff
    """
    if full_cn is None:
        return [None]
    if full_cn == 0:
        return [0]
    if full_cn == 1 and count_reg1 > min_read:
        return [1]
    prob = []
    nsum = count_reg1 + count_reg2
    if nsum == 0:
        return [None]
    for i in range(full_cn + 1):
        depthexpected = float(nsum) * float(i) / float(full_cn)
        if i == 0:
            depthexpected = (ERROR_RATE / 3) * float(nsum)
        if i == full_cn:
            depthexpected = float(nsum) - ERROR_RATE * float(nsum)
        if count_reg1 <= count_reg2:
            prob.append(poisson.pmf(int(count_reg1), depthexpected))
        else:
            prob.append(poisson.pmf(int(count_reg2), depthexpected))
    sum_prob = sum(prob)
    if sum_prob == 0:
        return [None]
    post_prob = [float(a) / float(sum_prob) for a in prob]
    if count_reg2 < count_reg1:
        post_prob = post_prob[::-1]
    post_prob_sorted = sorted(post_prob, reverse=True)
    if (
        post_prob.index(post_prob_sorted[0]) != 0
        and count_reg1 <= min_read
        and count_reg2 >= min_read
    ):
        return [0]
    if post_prob_sorted[0] >= POSTERIOR_CUTOFF_STRINGENT:
        return [post_prob.index(post_prob_sorted[0])]
    # output the two most likely scenarios
    cn_prob_filtered = [
        post_prob.index(post_prob_sorted[0]),
        round(post_prob_sorted[0], 3),
        post_prob.index(post_prob_sorted[1]),
        round(post_prob_sorted[1], 3),
    ]
    return cn_prob_filtered


def process_raw_call_gc(cn_prob, post_cutoff, keep_none=True):
    """
    Filter raw CN calls based on posterior probablity cutoff.
    For gene conversion cases, i.e. SNVs between paralogs.
    """
    cn_prob_filtered = []
    for cn_call in cn_prob:
        if len(cn_call) == 1:
            call_value = cn_call[0]
            if call_value is not None or keep_none:
                cn_prob_filtered.append(call_value)
        elif cn_call[1] > post_cutoff:
            cn_prob_filtered.append(cn_call[0])
        elif keep_none:
            cn_prob_filtered.append(None)
    return cn_prob_filtered


def process_raw_call_denovo(
    cn_prob, post_cutoff1, post_cutoff2, list_total_cn=None, keep_none=True
):
    """
    Filter raw CN calls based on posterior probablity cutoff.
    For de novel variant calling, i.e. non-gene-conversion cases.
    For less confident calls that are not copy number zero,
    return the smaller CN call.
    Also keep the variant if called CN is equal to total CN at the site.
    This makes sure a variant is always kept if CN>=1
    but we can't distinguish 1 vs 2.
    """
    cn_prob_filtered = []
    for i, cn_call in enumerate(cn_prob):
        if len(cn_call) == 1:
            call_value = cn_call[0]
            if call_value is not None or keep_none:
                cn_prob_filtered.append(call_value)
        else:
            if list_total_cn is not None:
                total_cn = list_total_cn[i]
                keep_var = (cn_call[0] > 0 and cn_call[2] > 0) or (
                    cn_call[0] == total_cn or cn_call[2] == total_cn
                )
            else:
                keep_var = cn_call[0] > 0 and cn_call[2] > 0
            if cn_call[1] > post_cutoff1:
                cn_prob_filtered.append(cn_call[0])
            elif keep_var:
                if cn_call[1] > post_cutoff2:
                    cn_prob_filtered.append(cn_call[0])
                else:
                    cn_prob_filtered.append(min(cn_call[0], cn_call[2]))
            elif keep_none:
                cn_prob_filtered.append(None)
    return cn_prob_filtered


def call_cn_snp(total_cn, lsnp1, lsnp2, threshold=0.6):
    """
    Call CN for SNP sites between geneA and geneB.
    Use a loose cutoff as this is for CNV/hybrid group calling.
    """
    cn_prob = []
    for i, count1 in enumerate(lsnp1):
        count2 = lsnp2[i]
        cn_prob.append(call_reg1_cn(total_cn, count1, count2))
    cn_call = process_raw_call_gc(cn_prob, threshold)
    return cn_call


def call_cn_var_homo(total_cn, lsnp1, lsnp2, max_cn=None, min_read=4):
    """
    Call CN for variant sites in homology regions.
    """
    cn_prob = []
    for i, count1 in enumerate(lsnp1):
        count2 = lsnp2[i]
        cn_prob.append(call_reg1_cn(total_cn, count1, count2, min_read))
    cn_call = []
    for site_call in process_raw_call_denovo(cn_prob, 0.8, 0.65):
        if site_call is None:
            cn_call.append(None)
        else:
            if max_cn is None:
                cn_call.append(min(site_call, total_cn - 2))
            else:
                cn_call.append(min(site_call, max_cn))
    return cn_call


def call_cn_var(total_cn, lsnp1, lsnp2, min_read=2):
    """
    Call CN for variant sites in non-homology regions.
    """
    cn_prob = []
    for i, count1 in enumerate(lsnp1):
        count2 = lsnp2[i]
        cn_prob.append(call_reg1_cn(total_cn, count1, count2, min_read))
    cn_call = process_raw_call_gc(cn_prob, 0.8)
    return cn_call


def get_called_variants(var_list, cn_prob_processed, n=0):
    """
    Return called variants based on called copy number and list of variant names
    """
    total_callset = []
    if n != 0:
        assert len(var_list) == len(cn_prob_processed) + n
    for i, cn_called in enumerate(cn_prob_processed):
        if cn_called is not None and cn_called != 0:
            for _ in range(cn_called):
                total_callset.append(var_list[i+n])
    return total_callset



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import os
import sys
import numpy as np
from collections import namedtuple
dir_name = os.path.dirname(os.path.dirname(__file__))
sys.path.append(dir_name)

from ..depth_calling.utilities import open_alignment_file
from ..depth_calling.gene_caller import GeneCaller
from .phasing_gba import PhasingGba


class GbaCaller(GeneCaller):
    sample_call = namedtuple(
        "sample_call",
        "Coverage_MAD Median_depth deletion_CN deletion_CN_raw \
        GBAP1_like_variant_exon9_11 other_variants variant_raw_count \
        snp_call snp_raw haplotypes",
    )

    def __init__(self):
        super().__init__()

    def improve_total_cn(self):
        """
        Improve GBA total cn when cn is high
        """
        # there could be noise in high CN calls. pick the most likely total CN
        # so that sites at the beginning of GBA (should always be CN(GBA)=2)
        # are called close to CN2
        self.total_cn = None
        if self.gmm_call.cn is not None and self.gmm_call.cn <= 4:
            self.total_cn = self.gmm_call.cn + 2
        elif self.gmm_call.depth_value > 4.25:
            total_cn_low = int(np.floor(self.gmm_call.depth_value + 2))
            total_cn_high = int(np.ceil(self.gmm_call.depth_value + 2))
            gene_sites_low = [a*total_cn_low for a in self.gene1_fraction[-20:]]
            gene_sites_high = [a*total_cn_high for a in self.gene1_fraction[-20:]]
            # substract the expected CN, which is 2
            gene_sites_low_diff = abs(np.median([a - 2 for a in gene_sites_low]))
            gene_sites_high_diff = abs(np.median([a - 2 for a in gene_sites_high]))
            if gene_sites_low_diff < 0.1 and gene_sites_high_diff > 0.2:
                self.total_cn = total_cn_low
            elif gene_sites_high_diff < 0.1 and gene_sites_low_diff > 0.2:
                self.total_cn = total_cn_high
        if self.total_cn is None and self.gmm_call.cn is not None:
            self.total_cn = self.gmm_call.cn + 2

    def call_gba_recombinants(self, debug=False):
        """
        Call variants in homology region by analyzing haplotypes
        """
        phase_gba = PhasingGba()
        bam_handle = open_alignment_file(self.bamfile, self.reference)
        phase_gba.set_par(
            bam_handle,
            self.call_parameters.haplotype_db["Exon9-11"],
            self.total_cn,
            self.reference
        )
        bam_handle.close()
        return phase_gba.call_variants_from_haplotype(debug=debug)

    def call(self):
        """Perform variant calling"""
        # 1. read counting, normalization
        self.normalize_depth()
        if self.normalized_depth.normalized["GBAdel"] is None:
            return self.sample_call(
                self.normalized_depth.mad,
                self.normalized_depth.mediandepth,
                None, None, None, None, None, None, None, None)
        # 2. call copy number
        self.call_cn("GBAdel")
        # 3. Get read counts and SNP ratios
        self.count_reads_diff_sites_and_variants()
        self.improve_total_cn()
        if self.total_cn is None:
            return self.sample_call(
                self.normalized_depth.mad,
                self.normalized_depth.mediandepth,
                self.gmm_call.cn, self.gmm_call.depth_value,
                None, None, self.raw_count, None, None, None
            )
        # 4. Call variants (simple)
        self.call_cn_diff_sites_and_variants()
        # site 48 onwards gets into GBA gene.
        # Mark here so we can easily visualize breakpoint location
        self.raw_gene1_cn_out = self.raw_gene1_cn
        self.raw_gene1_cn_out.insert(48, 'gene')
        # 5. Call variants (use the phasing approach in homology region Exon9-11)
        recombinant_variants = []
        variants_called_from_haplotype = self.call_gba_recombinants()
        if (
            variants_called_from_haplotype.is_biallelic or
            variants_called_from_haplotype.is_carrier
        ):
            recombinant_variants.append(
                ",".join(variants_called_from_haplotype.variants_allele1)
                )
            recombinant_variants.append(
                ",".join(variants_called_from_haplotype.variants_allele2)
                )
        # 6. Return final call
        self.cn_call_snp_out = self.cn_call_snp
        self.cn_call_snp_out.insert(48, 'gene')
        sample_gba_call = self.sample_call(
            self.normalized_depth.mad,
            self.normalized_depth.mediandepth,
            self.gmm_call.cn,
            self.gmm_call.depth_value,
            "/".join(recombinant_variants),
            ",".join(self.total_callset),
            self.raw_count,
            ",".join([str(a) for a in self.cn_call_snp_out]),
            ",".join([str(a) for a in self.raw_gene1_cn_out]),
            variants_called_from_haplotype._asdict()
        )
        return sample_gba_call



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from collections import namedtuple, OrderedDict
import multiprocessing as mp
from functools import partial
import numpy as np
from statsmodels.nonparametric.smoothers_lowess import lowess
from .utilities import open_alignment_file


MAD_CONSTANT = 1.4826


def get_normed_depth(bamf, region_dic, nCores=1, reference=None, gc_correct=True):
    """
    Return the normalized depth values and coverage stats for a sample
    given a bam file
    """
    counts_for_normalization, gc_for_normalization, region_type_cn, read_length = count_reads_and_prepare_for_normalization(
        bamf, region_dic, nCores, reference
    )
    normed_depth = normalize(
        counts_for_normalization,
        gc_for_normalization,
        region_type_cn,
        read_length,
        gc_correct,
    )

    return normed_depth


def get_normed_depth_from_count(count_file, region_dic, read_length, gc_correct=True):
    """
    Return the normalized depth values and coverage stats for a sample from
    a count file.
    """
    count_dic = get_count_from_file(count_file)
    counts_for_normalization, gc_for_normalization, region_type_cn = process_counts_and_prepare_for_normalization(
        count_dic, region_dic
    )
    normed_depth = normalize(
        counts_for_normalization,
        gc_for_normalization,
        region_type_cn,
        read_length,
        gc_correct,
    )

    return normed_depth


def median_correction(counts):
    """Return values corrected by median."""
    y_counts = np.array(counts)
    y_counts = y_counts / np.median(y_counts)
    return y_counts


def gc_correction(counts, gc, scale_coefficient=0.9):
    """Return values corrected by GC content."""
    y_counts = np.array(counts)
    x_gc = np.array(gc)
    y_counts = y_counts / np.median(y_counts)
    value_lowess = lowess(y_counts, x_gc, return_sorted=False)
    sample_median = np.median(y_counts)
    gc_corrected = []
    for i in range(len(y_counts)):
        scale_factor = scale_coefficient * min(y_counts[i], 2)
        gc_corrected.append(
            y_counts[i] + scale_factor * (sample_median - value_lowess[i])
        )

    return gc_corrected


def get_read_count(bamfile, region, mapq_cutoff=0):
    """
    Return the number of reads that align to a region.
    Keep duplicate reads.
    Keep unmapped reads with mapped mates.
    """
    reads = bamfile.fetch(region[0], region[1], region[2])
    nreads = 0
    for read in reads:
        if (
            read.mapq >= mapq_cutoff
            and read.is_secondary == 0
            and read.is_supplementary == 0
            and read.reference_start >= region[1]
            and read.reference_start < region[2]
        ):
            nreads += 1
    return nreads


def mad(list_of_number):
    """Return the median of absolute deviation."""
    med = np.median(list_of_number)
    return MAD_CONSTANT * np.median([abs(a - med) for a in list_of_number])


def normalize(
    counts_for_normalization,
    gc_for_normalization,
    region_type_cn,
    read_length,
    gc_correct,
):
    """
    Return the normalized depth values for a sample.
    Median normalization and/or GC normalization
    """
    gc_corrected_depth = gc_correction(counts_for_normalization, gc_for_normalization)
    if gc_correct is True:
        # GC normalization
        corrected_depth = gc_corrected_depth
    else:
        # Median normalization
        corrected_depth = median_correction(counts_for_normalization)

    vmedian = np.median(counts_for_normalization) * read_length
    vmad = round(mad(gc_corrected_depth) / np.median(gc_corrected_depth), 3)

    # Also return median depth of this sample and the coverage MAD.
    norm_count = {}
    for i, (region_type, hap_cn) in enumerate(region_type_cn.items()):
        if vmedian == 0:
            norm_count.setdefault(region_type, None)
        else:
            norm_count.setdefault(region_type, 2 * hap_cn * corrected_depth[i])
    depth_value = namedtuple("depth_value", "normalized mediandepth mad")
    normalized_bin = depth_value(norm_count, vmedian, vmad)
    return normalized_bin


def get_read_length(reads, number_to_count=2000):
    """Return the median read length from a set of reads."""
    counter = 0
    read_length = []
    for read in reads:
        counter += 1
        read_length.append(read.query_length)
        if counter > number_to_count:
            break
    return np.median(read_length)


def count_reads_and_prepare_for_normalization(
    bamf, region_dic, nCores=1, reference=None
):
    """
    Return the normalized depth values and coverage stats for a sample
    given a bam file
    """
    bamfile = open_alignment_file(bamf, reference)
    # Store read counts in each interval
    counts_for_normalization = []
    gc_for_normalization = []
    region_type_cn = OrderedDict()
    for region_type in region_dic:
        if region_type not in ["norm", "sex"]:
            lcount = []
            region_length = None
            hap_cn = None
            for (region, gc) in region_dic[region_type]:
                region_reads = get_read_count(bamfile, region)
                lcount.append(region_reads)
                if "_hapcn" in region[3]:
                    region_length = region[2] - region[1]
                    gc_for_normalization.append(float(gc))
                    hap_cn = int(region[3].split("hapcn")[1])
            if region_length is None or hap_cn is None:
                raise Exception("Problem with region definition. Length not specified.")
            count_sum = sum(lcount)
            counts_for_normalization.append(count_sum / (hap_cn * region_length))
            region_type_cn.setdefault(region_type, hap_cn)

    # Get read length from the last region
    reads = bamfile.fetch(region[0], region[1], region[2])
    read_length = get_read_length(reads)

    lregion = [
        (region[0], region[1], region[2], gc) for (region, gc) in region_dic["norm"]
    ]
    if nCores == 1:
        for region_out in get_normalization_region_values(lregion, bamf, reference):
            counts_for_normalization.append(region_out[0])
            gc_for_normalization.append(region_out[1])
    else:
        get_normed_depth_bam = partial(
            get_normalization_region_values, bam=bamf, reference=reference
        )
        region_groups = partition(lregion, nCores)
        pool = mp.Pool(nCores)
        result = pool.map(get_normed_depth_bam, region_groups)
        pool.close()
        for result_group in result:
            for region_out in result_group:
                counts_for_normalization.append(region_out[0])
                gc_for_normalization.append(region_out[1])

    bamfile.close()

    return counts_for_normalization, gc_for_normalization, region_type_cn, read_length


def partition(lst, n):
    """Partion a list"""
    return [lst[i::n] for i in range(n)]


def get_normalization_region_values(l, bam, reference=None):
    """Perform read counting in a list of regions."""
    bamfile = open_alignment_file(bam, reference)
    lcount = []
    for region in l:
        num_reads = get_read_count(bamfile, region)
        region_length = int(region[2]) - int(region[1])
        norm_depth = num_reads / region_length
        region_gc = float(region[-1])
        lcount.append((norm_depth, region_gc))
    bamfile.close()
    return lcount


def get_count_from_file(count_file):
    """Parse count file"""
    count_dic = {}
    with open(count_file) as f:
        for line in f:
            at = line.strip().split()
            count_dic.setdefault(at[3], int(at[-1]))
    return count_dic


def process_counts_and_prepare_for_normalization(count_dic, region_dic):
    """
    Return the normalized depth values and coverage stats for a sample from
    a count dictionary.
    """
    counts_for_normalization = []
    gc_for_normalization = []
    region_type_cn = OrderedDict()
    for region_type in region_dic:
        if region_type not in ["norm", "sex"]:
            lcount = []
            region_length = None
            hap_cn = None
            for (region, gc) in region_dic[region_type]:
                lcount.append(count_dic[region[3]])
                if "_hapcn" in region[3]:
                    region_length = region[2] - region[1]
                    gc_for_normalization.append(float(gc))
                    hap_cn = float(region[3].split("hapcn")[1])
            if region_length is None or hap_cn is None:
                raise Exception("Problem with region definition. Length not specified.")
            count_sum = sum(lcount)
            counts_for_normalization.append(count_sum / (hap_cn * region_length))
            region_type_cn.setdefault(region_type, hap_cn)

    for (region, gc) in region_dic["norm"]:
        region_length = int(region[2]) - int(region[1])
        counts_for_normalization.append(count_dic[region[3]] / region_length)
        gc_for_normalization.append(gc)

    return counts_for_normalization, gc_for_normalization, region_type_cn


def count_reads_and_call_sex(
    bamf, region_dic, nCores=1, reference=None
):
    """
    Return the sex of a sample given a bam file
    """
    bamfile = open_alignment_file(bamf, reference)
    # Store read counts in each interval
    x_count = []
    y_count = []
    auto_count = []
    for (region, gc) in region_dic["sex"]:
        region_length = region[2] - region[1]
        region_reads = get_read_count(bamfile, region)
        if "X" in region[0]:
            x_count.append(region_reads/region_length)
        if "Y" in region[0]:
            y_count.append(region_reads/region_length)
    i = 0
    for (region, gc) in region_dic["norm"]:
        i += 1
        if i <= 100:
            region_length = region[2] - region[1]
            region_reads = get_read_count(bamfile, region)
            auto_count.append(region_reads/region_length)
        else:
            break
    bamfile.close()

    # enum
    sex = None
    if np.median(y_count)/np.median(auto_count) < 0.01 and np.median(x_count)/np.median(auto_count) > 0.65:
        return "female"
    if np.median(y_count)/np.median(auto_count) > 0.05 and np.median(x_count)/np.median(auto_count) < 0.65:
        return "male"

    return sex



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from collections import namedtuple
import numpy as np
from scipy.stats import norm

# sd for CN=0 is arbitrarily set at 0.032
SIGMA_CN0 = 0.032
PV_CUTOFF = 1e-3
POSTERIOR_CUTOFF = 0.95
# CN = 0 - 10
DEFAULT_GMM_NSTATE = 11


class Gmm:
    def __init__(self, num_state=DEFAULT_GMM_NSTATE):
        self.mu_state = []
        self.sigma_state = []
        self.prior_state = []
        self.value_shift = None
        self.nstate = num_state

    def set_gmm_par(self, dpar_tmp, svid):
        """Return the complete set of gmm parameters based on depth value."""
        if svid not in dpar_tmp:
            raise Exception("Variant id %s is not recognized." % svid)
        gmm_parameter = dpar_tmp[svid]
        # The gmm parameter file stores the adjustment factor (value_shift) for
        # all depth values. This value should be one if there is no bias
        # between target regions and the normalization regions.
        # But this value may vary slightly depending on the aligner
        # choice and aligner setting. If running the caller on a population,
        # it may be desirable to run GMM modeling for the population and make
        # this value more accurate.
        self.value_shift = float(gmm_parameter["shift"][0])

        # The means are modeled as 0, 0.5, 1, 1.5, 2, 2.5, etc.
        # (We could also do 0, 1, 2, 3, 4, 5, etc.)
        # The gmm parameter file stores the mean depth values for CN=2 and CN=3
        self.mu_state = [
            0,
            0.5,
            float(gmm_parameter["mean"][0]),
            float(gmm_parameter["mean"][1]),
        ]
        mu_width = float(gmm_parameter["mean"][1]) - float(gmm_parameter["mean"][0])
        for i in range(self.nstate):
            if i > 3:
                self.mu_state.append(1 + mu_width * (i - 2))

        sum_prior = sum([float(a) for a in gmm_parameter["prior"]])
        if sum_prior >= 1:
            raise Exception("Sum of priors is larger than 1.")
        for i in range(self.nstate):
            # The gmm parameter file stores the prior frequencies for CN=0-6
            if i < len(gmm_parameter["prior"]):
                self.prior_state.append(float(gmm_parameter["prior"][i]))
            else:
                prior_value = (1 - sum_prior) / (
                    self.nstate - len(gmm_parameter["prior"])
                )
                self.prior_state.append(prior_value)

        self.sigma_state = [SIGMA_CN0]
        # The gmm parameter file stores the standard deviation for CN=2.
        # The other sd values are derived from it.
        for i in range(self.nstate):
            if i > 0:
                sigma_value = float(gmm_parameter["sd"][0]) * np.sqrt(float(i) / 2)
                self.sigma_state.append(sigma_value)

    def gmm_call(self, val):
        """Return the final copy number call."""
        val_new = (val / 2) / self.value_shift
        fcall = self.call_post_prob(val_new, POSTERIOR_CUTOFF)
        if fcall is not None:
            gauss_p_value = self.get_gauss_pmf_cdf(
                val_new, self.mu_state[fcall], self.sigma_state[fcall]
            )[1]
            # apply another p-value cutoff
            # just comparing the depth value and the called CN
            if gauss_p_value < PV_CUTOFF:
                fcall = None
        cn_call = namedtuple("cn_call", "cn depth_value")
        return cn_call(fcall, round(val / self.value_shift, 3))

    def get_gauss_pmf_cdf(self, test_value, gauss_mean, gauss_sd):
        """Return the pmf and cdf of a gaussian distribution."""
        test_stats = (test_value - gauss_mean) / gauss_sd
        pdf = norm.pdf(test_stats) / gauss_sd
        p_value = min(norm.cdf(test_stats), 1 - norm.cdf(test_stats))
        return (pdf, p_value)

    def call_post_prob(self, val, post_cutoff):
        """Return the copy number call based on gaussian mixture model."""
        number_state = len(self.prior_state)
        prob = []
        for i in range(0, number_state):
            gauss_pmf = self.get_gauss_pmf_cdf(
                val, self.mu_state[i], self.sigma_state[i]
            )[0]
            prob.append(gauss_pmf * self.prior_state[i])
        sum_prob = float(sum(prob))
        post_prob = [float(a) / sum_prob for a in prob]
        max_prob = max(post_prob)
        if max_prob >= post_cutoff:
            return post_prob.index(max_prob)
        else:
            return None



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#


from collections import namedtuple


COMPLEMENT = {"A": "T", "T": "A", "C": "G", "G": "C", "N": "N"}


def reverse_complement(sequence):
    """Return the reverse complement of a sequence."""
    return "".join(COMPLEMENT[b] for b in sequence[::-1])


def get_nm(ltag):
    """Return the value of the NM tag."""
    for tag in ltag:
        if tag[0] == "NM":
            return tag[1]
    return None


def get_snp_position(pos_file, group=None):
    """Get all base differences listed in the SNP location file."""
    dsnp1 = {}
    dsnp2 = {}
    dindex = {}

    with open(pos_file) as read_pos:
        counter = -1
        for line in read_pos:
            if line[0] != "#" and line[0] != "\n":
                split_line = line.strip().split()
                if group is None or split_line[-1] == group:
                    counter += 1
                    reg1_name = split_line[1] + "_" + str(counter)
                    reg2_name = split_line[3] + "_" + str(counter)
                    reg1_base = split_line[2].upper()
                    reg2_base = split_line[4].upper()
                    if split_line[-1] != "-":
                        dsnp1.setdefault(reg1_name, "_".join([reg1_base, reg2_base]))
                        dsnp2.setdefault(reg2_name, "_".join([reg1_base, reg2_base]))
                    else:
                        dsnp1.setdefault(
                            reg1_name,
                            "_".join([reg1_base, reverse_complement(reg2_base)]),
                        )
                        dsnp2.setdefault(
                            reg2_name,
                            "_".join([reverse_complement(reg1_base), reg2_base]),
                        )
                    dindex.setdefault(reg1_name, counter)
                    dindex.setdefault(reg2_name, counter)
    if counter == -1:
        return None
    nchr = split_line[0]
    snp_lookup = namedtuple("snp_lookup", "dsnp1 dsnp2 nchr dindex")
    dbsnp = snp_lookup(dsnp1, dsnp2, nchr, dindex)
    return dbsnp


def passing_read(pileupread):
    """Return whether a read passes filter."""
    return (
        not pileupread.is_del
        and not pileupread.is_refskip
        and pileupread.alignment.is_secondary == 0
        and pileupread.alignment.is_supplementary == 0
        and pileupread.alignment.is_duplicate == 0
    )


def passing_read_stringent(pileupread):
    """Return whether a read passes more stringent filter."""
    number_mismatch = get_nm(pileupread.alignment.tags)
    align_len = pileupread.alignment.query_alignment_length
    read_len = len(pileupread.alignment.query_sequence)
    insert_size = pileupread.alignment.template_length
    # number_mismatch <= float(align_len) * 0.08
    # and pileupread.query_position > 0
    # and pileupread.query_position < read_len - 1
    return abs(insert_size) < 1000


def get_reads_by_region(
    bamfile_handle, nchr, dsnp, dindex, min_mapq=0, stringent=False
):
    """
    Return the number of reads supporting region1 and region2, forward and reverse.
    """
    lsnp1_forward = []
    lsnp1_reverse = []
    lsnp2_forward = []
    lsnp2_reverse = []
    for _ in dsnp:
        lsnp1_forward.append(set())
        lsnp1_reverse.append(set())
        lsnp2_forward.append(set())
        lsnp2_reverse.append(set())

    for snp_position_ori in dsnp:
        snp_position = int(snp_position_ori.split("_")[0])
        for pileupcolumn in bamfile_handle.pileup(
            nchr,
            snp_position - 1,
            snp_position + 1,
            truncate=True,
            stepper="nofilter",
            ignore_overlaps=False,
            ignore_orphan=False,
        ):
            site_position = pileupcolumn.pos + 1
            if site_position == snp_position:
                reg1_allele, reg2_allele = dsnp[snp_position_ori].split("_")
                for read in pileupcolumn.pileups:
                    if (
                        passing_read(read)
                        and read.alignment.mapping_quality >= min_mapq
                    ):
                        dsnp_index = dindex[snp_position_ori]
                        read_name = read.alignment.query_name
                        read_seq = read.alignment.query_sequence
                        if stringent is False or passing_read_stringent(read):
                            reg1_allele_split = reg1_allele.split(",")
                            reg2_allele_split = reg2_allele.split(",")
                            start_pos = read.query_position
                            for allele in reg1_allele_split:
                                end_pos = start_pos + len(allele)
                                if read_seq[start_pos:end_pos] == allele:
                                    if read.alignment.is_reverse:
                                        lsnp1_reverse[dsnp_index].add(read_name)
                                    else:
                                        lsnp1_forward[dsnp_index].add(read_name)
                            for allele in reg2_allele_split:
                                end_pos = start_pos + len(allele)
                                if read_seq[start_pos:end_pos] == allele:
                                    if read.alignment.is_reverse:
                                        lsnp2_reverse[dsnp_index].add(read_name)
                                    else:
                                        lsnp2_forward[dsnp_index].add(read_name)
    return lsnp1_forward, lsnp1_reverse, lsnp2_forward, lsnp2_reverse


def get_fraction(lsnp1, lsnp2):
    """Return the fraction of reads supporting region1."""
    reg1_fraction = []
    for index in range(len(lsnp1)):
        sumdepth = lsnp1[index] + lsnp2[index]
        if sumdepth == 0:
            reg1_fraction.append(0)
        else:
            reg1_fraction.append(float(lsnp1[index]) / float(sumdepth))
    return reg1_fraction


def merge_reads(list_to_merge):
    """
    Merge sets of reads (forward/reverse, region1/region2)
    """
    merged_reads = []
    for i in range(len(list_to_merge[0])):
        reads_per_site = set()
        for lsnp in list_to_merge:
            reads_per_site = reads_per_site.union(lsnp[i])
        merged_reads.append(reads_per_site)
    return merged_reads


def get_supporting_reads(bamfile_handle, dsnp1, dsnp2, nchr, dindex):
    """
    Return the number of supporting reads at each position in
    both region1 and region2.
    """
    assert len(dsnp1) == len(dsnp2)
    # Go through SNP sites in both regions,
    # and count the number of reads supporting each gene.
    lsnp1_reg1_for, lsnp1_reg1_rev, lsnp2_reg1_for, lsnp2_reg1_rev = get_reads_by_region(
        bamfile_handle, nchr, dsnp1, dindex
    )
    lsnp1_reg2_for, lsnp1_reg2_rev, lsnp2_reg2_for, lsnp2_reg2_rev = get_reads_by_region(
        bamfile_handle, nchr, dsnp2, dindex
    )
    lsnp1 = merge_reads(
        [lsnp1_reg1_for, lsnp1_reg1_rev, lsnp1_reg2_for, lsnp1_reg2_rev]
    )
    lsnp2 = merge_reads(
        [lsnp2_reg1_for, lsnp2_reg1_rev, lsnp2_reg2_for, lsnp2_reg2_rev]
    )
    return [len(a) for a in lsnp1], [len(a) for a in lsnp2]


def get_supporting_reads_single_region(bamfile_handle, dsnp1, nchr, dindex):
    """
    Return the number of supporting reads at each position only in region1, as well
    as the number of alt reads in forward and reverse.
    """
    lsnp1_for, lsnp1_rev, lsnp2_for, lsnp2_rev = get_reads_by_region(
        bamfile_handle, nchr, dsnp1, dindex, 10
    )
    lsnp1 = merge_reads([lsnp1_for, lsnp1_rev])
    lsnp2 = merge_reads([lsnp2_for, lsnp2_rev])
    return (
        [len(a) for a in lsnp1],
        [len(a) for a in lsnp2],
        [len(a) for a in lsnp1_for],
        [len(a) for a in lsnp1_rev],
    )



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import os
import sys
from collections import namedtuple
from pprint import pprint
dir_name = os.path.dirname(os.path.dirname(__file__))
sys.path.append(dir_name)
from ..depth_calling.haplotype import (
    extract_hap,
    extend_hap_5p,
    extend_hap_3p,
    group_haps,
    filter_hap,
    join_subblocks,
)
from ..depth_calling.phasing import Phasing


class PhasingGba(Phasing):
    assembled_haplotypes = namedtuple(
        "assembled_haplotypes", "full_haplotypes hap_5p hap_3p"
    )
    def __init__(self):
        super().__init__()
        self.variant_sites = [2, 3, 6, 7]
        self.trusted_hap_threshold = [
            "1111112111",
            "1121111111",
            "1112111111",
            "1111111221"
            ]
        self.trusted_hap_switch_point = [
            "1111112111",
            "1121111111",
            "1112111111",
            "1111111221"
            ]
        self.variant_names = [
            "A495P",
            "L483P",
            "D448H",
            "c.1263del",
        ]
        self.variant_names_in_one = ["c.1263del+RecTL"]
        # sites used to calculate depth/read counts
        self.flanking_site = {
            2: [[2, 3]],
            3: [[2, 3]],
            6: [[6, 7]],
            7: [[6, 7]]
            }

    def get_variants_gba(self, hap):
        """
        Get variant names given haplotype
        """
        variants = []
        for i in range(len(hap)):
            if hap[i] == '2' and i in self.variant_sites:
                variant_index = self.variant_sites.index(i)
                variants.append(self.variant_names[variant_index])
        if variants == ["A495P", "L483P"]:
            return ["RecNciI"]
        if variants == ["A495P", "L483P", "D448H"]:
            return ["RecTL"]
        if variants == ["A495P", "L483P", "D448H", "c.1263del"]:
            return ["c.1263del+RecTL"]
        return variants

    def assemble_haplotypes(self, debug=False):
        """
        Assembly haplotypes for exon9-11
        """
        # starting from left
        hap_count = extract_hap(self.haplotype_per_read, range(2, 4))
        hap_count = filter_hap(hap_count)
        haplotypes_to_extend = ["x" * 2 + a + "x" * 6 for a in hap_count]
        # Extend to 5p
        matching_haplotype_groups = group_haps(self.haplotype_per_read, haplotypes_to_extend)
        haplotypes_to_extend = extend_hap_5p(matching_haplotype_groups)
        matching_haplotype_groups = group_haps(self.haplotype_per_read, haplotypes_to_extend)
        haplotypes_to_extend = extend_hap_5p(matching_haplotype_groups)
        # Extend to 3p
        n = 0
        while n < 4:
            n += 1
            if haplotypes_to_extend != []:
                matching_haplotype_groups = group_haps(self.haplotype_per_read, haplotypes_to_extend)
                haplotypes_to_extend = extend_hap_3p(matching_haplotype_groups)
                if debug is True:
                    pprint(matching_haplotype_groups)
                    print(n, haplotypes_to_extend)
            else:
                break
        hap_5p = haplotypes_to_extend
        # starting from right
        hap_count = extract_hap(self.haplotype_per_read, [6, 7])
        hap_count = filter_hap(hap_count)
        haplotypes_to_extend = ["x" * 6 + a + a[-1] + "x" for a in hap_count]
        # Extend to 3p
        matching_haplotype_groups = group_haps(self.haplotype_per_read, haplotypes_to_extend)
        haplotypes_to_extend = extend_hap_3p(matching_haplotype_groups)
        # Extend to 5p
        n = 0
        while n < 4:
            n += 1
            if haplotypes_to_extend != []:
                matching_haplotype_groups = group_haps(self.haplotype_per_read, haplotypes_to_extend)
                haplotypes_to_extend = extend_hap_5p(matching_haplotype_groups)
                if debug is True:
                    pprint(matching_haplotype_groups)
                    print(n, haplotypes_to_extend)
            else:
                break
        hap_3p = haplotypes_to_extend
        full_hap, dcount = join_subblocks(self.haplotype_per_read, hap_5p, hap_3p, 2, 8)

        return self.assembled_haplotypes(full_hap, hap_5p, hap_3p)

    def check_deletion_bp_in_gene(self, var_index_in_haps):
        """In case of a deletion, check whether the deletion breakpoint is within the gene"""
        for hap in var_index_in_haps:
            # switching from 2s to 1s
            if (
                self.total_cn < 4 and
                hap.count("21") == 1 and
                "12" not in hap and
                "x" not in hap
            ):
                self.deletion_bp_in_gene = True

    def assess_hap(self, hap, var_index_in_haps, full_haplotypes):
        """
        Assess whether a haplotype is a variant haplotype that makes the
        sample a carrier, i.e. having only one copy of the WT gene
        """
        fully_assembled, var_sites_haps = self.assess_assembled_haps(full_haplotypes)
        good_var_indices = var_index_in_haps[hap]
        # Scenario 1: variant haplotype exists and there is only one copy of
        # the WT gene, based on depth at defined flanking sites
        # use different thresholds for trusted haplotypes
        if (hap in self.trusted_hap_threshold and len(full_haplotypes) <= 4):
            found_variant_in_short_hap, block_result = self.get_depth_for_blocks(
                good_var_indices, cn_threshold=self.CN_likelihood_threshold_loose)
        else:
            found_variant_in_short_hap, block_result = self.get_depth_for_blocks(
                good_var_indices, cn_threshold=self.CN_likelihood_threshold)
        if found_variant_in_short_hap:
            self.is_carrier = True
        # Scenario 2: considering all haplotypes, there is only one haplotype
        # that has the gene base at all variat sites
        # not based on depth. Only consider total_cn=3
        elif len(var_index_in_haps) == 1 and fully_assembled is True:
            carrier_from_haps = self.carrier_from_assembled_haps(var_sites_haps)
            if carrier_from_haps is True:
                self.is_carrier = True
        # Scenario 3: check switching points not included in flanking_site
        # for trusted haplotypes or all-2s to all-1s transitions
        if self.is_carrier is not True:
            if (
                hap in self.trusted_hap_switch_point or
                (hap.count("12") == 1 and "21" not in hap) or
                (hap.count("21") == 1 and "12" not in hap)
            ):
                for switching_hap in ["12", "21"]:
                    if switching_hap in hap:
                        switch_point = hap.index(switching_hap)
                        switch_block = self.get_depth_for_block(
                            [switch_point, switch_point + 1], switching_hap.index('1')
                            )
                        if switch_block not in block_result:
                            block_result.append(switch_block)
                        if (
                            switch_block.likelihood is not None and
                            switch_block.likelihood > self.CN_likelihood_threshold
                        ):
                            self.is_carrier = True
        if self.is_carrier is True:
            var_per_hap = self.var_call_per_haplotype(
                hap,
                self.get_variants_gba(hap),
                [a._asdict() for a in block_result],
                True
                )
        else:
            var_per_hap = self.var_call_per_haplotype(
                hap,
                self.get_variants_gba(hap),
                [a._asdict() for a in block_result],
                found_variant_in_short_hap
                )
        self.variants_details.append(var_per_hap)



#!/usr/bin/env python3
#
# Gauchian: GBA variant caller
# Copyright 2021 Illumina, Inc.
# All rights reserved.
#
# Author: Xiao Chen <xchen2@illumina.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import os
import argparse
import json
import logging
import datetime
from collections import namedtuple
from .caller.gba_caller import GbaCaller
from .depth_calling.utilities import (
    parse_gmm_file,
    parse_region_file,
)
from .depth_calling.snp_count import get_snp_position

MAD_THRESHOLD = 0.11
resource_info = namedtuple(
    "resource_info",
    "genome gmm_parameter region_dic snp_db var_db var_homo_db haplotype_db var_list",
)


def load_parameters():
    """Return parameters."""
    parser = argparse.ArgumentParser(
        description="WGS-based targeted variant caller for GBA."
    )
    parser.add_argument(
        "-m", "--manifest",
        help="Manifest listing absolute paths to input BAM/CRAM files",
        required=True,
    )
    parser.add_argument(
        "-g", "--genome", help="Reference genome, select from 19, 37, or 38", required=True
    )
    parser.add_argument("-o", "--outDir", help="Output directory", required=True)
    parser.add_argument("-p", "--prefix", help="Prefix to output file", required=True)
    parser.add_argument(
        "-t", "--threads",
        help="Number of threads to use. Default is 1",
        type=int,
        default=1,
        required=False,
    )
    parser.add_argument(
        "--reference",
        help="Optional path to reference fasta file for CRAM",
        required=False,
    )

    args = parser.parse_args()
    if args.genome not in ["19", "37", "38"]:
        raise Exception("Genome not recognized. Select from 19, 37, or 38")

    return args


def prepare_data_files(datadir, gene, genome):
    """Check data files and prepare resource data for GBA calling"""
    region_file = os.path.join(datadir, "%s_region_%s.bed" % (gene, genome))
    snp_file = os.path.join(datadir, "%s_SNP_%s.txt" % (gene, genome))
    gmm_file = os.path.join(datadir, "%s_gmm.txt" % gene)
    variant_file = os.path.join(
        datadir, "%s_target_variant_%s.txt" % (gene, genome)
    )
    variant_homology_file = os.path.join(
        datadir, "%s_target_variant_homology_region_%s.txt" % (gene, genome)
    )
    haplotype_file = os.path.join(datadir, "%s_haplotype_%s.txt" % (gene, genome))

    for required_file in [
        region_file,
        snp_file,
        variant_file,
        variant_homology_file,
        haplotype_file,
        gmm_file,
    ]:
        if os.path.exists(required_file) == 0:
            raise Exception("File %s not found." % required_file)

    region_dic = parse_region_file(region_file)

    snp_db = None
    if os.path.exists(snp_file):
        snp_db = get_snp_position(snp_file)

    var_list = []
    var_db = None
    if os.path.exists(variant_file):
        var_db = get_snp_position(variant_file)
        with open(variant_file) as f:
            for line in f:
                if line[0] != "#":
                    var_name = line.split()[-1]
                    var_list.append(var_name)

    var_homo_db = None
    if os.path.exists(variant_homology_file):
        var_homo_db = get_snp_position(variant_homology_file)
        with open(variant_homology_file) as f:
            for line in f:
                if line[0] != "#":
                    var_name = line.split()[-1]
                    var_list.append(var_name)

    HAPLOTYPE_VAR = set()
    haplotype_db = {}
    if os.path.exists(haplotype_file):
        with open(haplotype_file) as f:
            for line in f:
                at = line.split()
                HAPLOTYPE_VAR.add(at[-1])
        for variant in HAPLOTYPE_VAR:
            haplotype_db.setdefault(
                variant,
                get_snp_position(haplotype_file, variant)
                )

    gmm_parameter = None
    if os.path.exists(gmm_file):
        gmm_parameter = parse_gmm_file(gmm_file)

    call_parameters = resource_info(
        genome,
        gmm_parameter,
        region_dic,
        snp_db,
        var_db,
        var_homo_db,
        haplotype_db,
        var_list,
    )
    return call_parameters


def write_to_tsv(final_output, out_tsv):
    """Prepare tsv output"""
    header = [
        "Sample",
        "is_biallelic(GBAP1-like_variant_exon9-11)",
        "is_carrier(GBAP1-like_variant_exon9-11)",
        "CN(GBA+GBAP1)",
        "deletion_breakpoint_in_GBA",
        "GBAP1-like_variant_exon9-11",
        "other_unphased_variants"
        ]
    with open(out_tsv, "w") as tsv_output:
        tsv_output.write("\t".join(header) + "\n")
        for sample_id in final_output:
            final_call = final_output[sample_id]
            if final_call["haplotypes"] is not None:
                p1like_variants = None
                if final_call["GBAP1_like_variant_exon9_11"] != "":
                    p1like_variants = final_call["GBAP1_like_variant_exon9_11"]
                other_variants = None
                if final_call["other_variants"] != "":
                    other_variants = final_call["other_variants"]
                output_per_sample = [
                    sample_id,
                    final_call["haplotypes"]["is_biallelic"],
                    final_call["haplotypes"]["is_carrier"],
                    final_call["haplotypes"]["total_cn"],
                    final_call["haplotypes"]["deletion_bp_in_gene"],
                    p1like_variants,
                    other_variants
                ]
                if output_per_sample[4] is None:
                    output_per_sample[4] = "N/A"
                tsv_output.write("\t".join([str(a) for a in output_per_sample]) + "\n")
            else:
                tsv_output.write("\t".join([str(a) for a in [sample_id]+[None]*6]) + "\n")


def run():
    parameters = load_parameters()
    manifest = parameters.manifest
    outdir = parameters.outDir
    genome = parameters.genome
    prefix = parameters.prefix
    threads = parameters.threads
    reference_fasta = parameters.reference
    logging.basicConfig(level=logging.DEBUG)

    datadir = os.path.join(os.path.dirname(__file__), "data")
    if os.path.exists(outdir) == 0:
        os.makedirs(outdir)
    out_json = os.path.join(outdir, prefix + ".json")
    out_tsv = os.path.join(outdir, prefix + ".tsv")
    call_parameters = prepare_data_files(datadir, "GBA", genome)

    final_output = {}
    with open(manifest) as read_manifest:
        for line in read_manifest:
            bam_name = line.strip()
            sample_id = os.path.splitext(os.path.basename(bam_name))[0]
            if os.path.exists(bam_name) == 0:
                logging.warning(
                    "Input alignmet file for sample %s does not exist.", sample_id
                )
            else:
                logging.info(
                    "Processing sample %s at %s", sample_id, datetime.datetime.now()
                )
                gba_caller = GbaCaller()
                gba_caller.set_par(
                    bam_name,
                    call_parameters,
                    threads,
                    reference_fasta,
                )
                gba_call = gba_caller.call()._asdict()
                # Use normalized coverage MAD across stable regions
                # as a sample QC measure.
                if gba_call["Coverage_MAD"] > MAD_THRESHOLD:
                    logging.warning(
                        "Sample %s has uneven coverage. CN calls may be unreliable.",
                        sample_id,
                    )
                final_output.setdefault(sample_id, gba_call)

    # Write to json
    logging.info("Writing to json at %s", datetime.datetime.now())
    with open(out_json, "w") as json_output:
        json.dump(final_output, json_output)

    # Write to tsv
    logging.info('Writing to tsv at %s', datetime.datetime.now())
    write_to_tsv(final_output, out_tsv)


if __name__ == "__main__":
    run()



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='gauchian',
      version='1.0.2',
      description='WGS-based GBA variant caller',
      long_description=readme(),
      long_description_content_type="text/markdown",
      classifiers=[],
      keywords='GBA',
      url='https://github.com/illumina/Gauchian',
      author='Xiao Chen',
      author_email='xchen2@illumina.com',
      license='GPLv3',
      packages=['gauchian', 'gauchian.caller', 'gauchian.depth_calling'],
      package_data={'gauchian': ['data/*']},
      install_requires=['pysam', 'numpy', 'scipy', 'statsmodels'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      entry_points={
          'console_scripts': [
              'gauchian=gauchian.gauchian:run'
              ]
      },
      include_package_data=True,
      zip_safe=False)
